exception Error of string
let error msg = raise (Error msg)


let lex file =
    let fd          = try open_in file
                      with Sys_error(msg) -> error msg      in
    let finally ()  = close_in fd                           in
    let lexbuf      = Lexing.from_channel fd                in
    let rec loop lb =
        match Scan.token lb with
            | Parse.EOF -> ()
            | x         -> ( print_endline (Scan.tok2str x)
                           ; loop lb
                           )
    in try
        ( loop lexbuf
        ; finally ()
        )
    with e ->
        ( finally ()
        ; raise e
        )

let open_err f = try open_in f with Sys_error(msg) -> error msg

let parse fd = 
    let finally ()  = close_in fd                           in
    let lexbuf      = Lexing.from_channel fd                in
    try 
        let tree = Parse.document Scan.token lexbuf in
            ( finally ()
            ; tree
            )
    with 
    | Parsing.Parse_error -> 
        ( finally ()
        ; error (Printf.sprintf "parse error around character %d" 
                (Lexing.lexeme_start lexbuf))
        )         
    | e ->
        ( finally ()
        ; raise e
        )

(*
let search d best = Discrepancy.disc_search Choose.rep (Convert.initial_state d) best
*)
  (* useful only for testing the alarm clock *)
(*
let search d best = Discrepancy.disc_search Choose.one (Convert.initial_state d) best
*)
let search d best = Discrepancy.disc_search Generate.generate (Convert.initial_state d) best
  (* useful only for debugging the search engine *)

(* let search d best = best (* until generate gets done *) *)

module S = Tree.Solution

let finish time ast =
  let _ = Sys.signal Sys.sigalrm (Sys.Signal_handle Tree.make_time_expire) in
  let time = int_of_string time in
  let time = if time > 3 then time - 3 else 0 in
  let _ = if time > 0 then ignore (Unix.alarm time) else Tree.make_time_expire () in
  let trivial = Emit.from_ast (Convert.compress_ast ast) in
  let orig = List.rev trivial.S.emitted in
try (
  let doco = Convert.compress ast in
  let res = search doco trivial in
(*
  let _ = prerr_string ("We found: " ^ string_of_int res.S.cost ^ "\n") in
  let _ = prerr_string ("Trivial found: " ^ string_of_int trivial.S.cost ^ "\n") in
*)
  let res' = List.rev res.S.emitted in
  let res'' = if Tree.cmp_document (Convert.emits_to_tree_doc orig)
                                   (Convert.emits_to_tree_doc res')
                 then res'
                 else (prerr_string "Bad Optimization\n"; orig) in
  Emit.write stdout res'') with _ -> (Emit.write stdout orig)

(* The time limit will depend on the input file and it will be given
to your program as a number of seconds, both in its first command-line
argument and in the TLIMIT environment variable.  *)
  
let main () =
    let argv        = Array.to_list Sys.argv in
    let this        = Filename.basename (List.hd argv) in
    match List.tl argv with
        | [time]       -> finish time (parse stdin)
        | [time; file] -> finish time (parse (open_err file))
        | _  -> error (Printf.sprintf "usage: %s seconds [inputfile]" this)
        

let _ = main ()
let _ = exit 0


